package org.apache.commons.lang3.time;

package .tmp;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.text.ParseException;
import java.text.ParsePosition;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Locale;
import java.util.Map;
import java.util.TimeZone;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import org.apache.commons.lang3.LocaleUtils;
import org.apache.commons.lang3.time.DateParser;

public class FastDateParser implements DateParser, Serializable {
  private static final long serialVersionUID = 3L;
  
  static final Locale JAPANESE_IMPERIAL = new Locale("ja", "JP", "JP");
  
  private final String pattern;
  
  private final TimeZone timeZone;
  
  private final Locale locale;
  
  private final int century;
  
  private final int startYear;
  
  private transient List<StrategyAndWidth> patterns;
  
  private static final Comparator<String> LONGER_FIRST_LOWERCASE = Comparator.reverseOrder();
  
  protected FastDateParser(String pattern, TimeZone timeZone, Locale locale) {
    this(pattern, timeZone, locale, null);
  }
  
  protected FastDateParser(String pattern, TimeZone timeZone, Locale locale, Date centuryStart) {
    int centuryStartYear;
    this.pattern = pattern;
    this.timeZone = timeZone;
    this.locale = LocaleUtils.toLocale(locale);
    Calendar definingCalendar = Calendar.getInstance(timeZone, this.locale);
    if (centuryStart != null) {
      definingCalendar.setTime(centuryStart);
      centuryStartYear = definingCalendar.get(1);
    } else if (this.locale.equals(JAPANESE_IMPERIAL)) {
      centuryStartYear = 0;
    } else {
      definingCalendar.setTime(new Date());
      centuryStartYear = definingCalendar.get(1) - 80;
    } 
    this.century = centuryStartYear / 100 * 100;
    this.startYear = centuryStartYear - this.century;
    init(definingCalendar);
  }
  
  private void init(Calendar definingCalendar) {
    this.patterns = new ArrayList<>();
    StrategyParser fm = new StrategyParser(this, definingCalendar);
    while (true) {
      StrategyAndWidth field = fm.getNextStrategy();
      if (field == null)
        break; 
      this.patterns.add(field);
    } 
  }
  
  private static boolean isFormatLetter(char c) {
    return ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z'));
  }
  
  public String getPattern() {
    return this.pattern;
  }
  
  public TimeZone getTimeZone() {
    return this.timeZone;
  }
  
  public Locale getLocale() {
    return this.locale;
  }
  
  public boolean equals(Object obj) {
    if (!(obj instanceof org.apache.commons.lang3.time.FastDateParser))
      return false; 
    org.apache.commons.lang3.time.FastDateParser other = (org.apache.commons.lang3.time.FastDateParser)obj;
    return (this.pattern.equals(other.pattern) && this.timeZone.equals(other.timeZone) && this.locale.equals(other.locale));
  }
  
  public int hashCode() {
    return this.pattern.hashCode() + 13 * (this.timeZone.hashCode() + 13 * this.locale.hashCode());
  }
  
  public String toString() {
    return "FastDateParser[" + this.pattern + ", " + this.locale + ", " + this.timeZone.getID() + "]";
  }
  
  public String toStringAll() {
    return "FastDateParser [pattern=" + this.pattern + ", timeZone=" + this.timeZone + ", locale=" + this.locale + ", century=" + this.century + ", startYear=" + this.startYear + ", patterns=" + this.patterns + "]";
  }
  
  private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    Calendar definingCalendar = Calendar.getInstance(this.timeZone, this.locale);
    init(definingCalendar);
  }
  
  public Object parseObject(String source) throws ParseException {
    return parse(source);
  }
  
  public Date parse(String source) throws ParseException {
    ParsePosition pp = new ParsePosition(0);
    Date date = parse(source, pp);
    if (date == null) {
      if (this.locale.equals(JAPANESE_IMPERIAL))
        throw new ParseException("(The " + this.locale + " locale does not support dates before 1868 AD)\nUnparseable date: \"" + source, pp
            .getErrorIndex()); 
      throw new ParseException("Unparseable date: " + source, pp.getErrorIndex());
    } 
    return date;
  }
  
  public Object parseObject(String source, ParsePosition pos) {
    return parse(source, pos);
  }
  
  public Date parse(String source, ParsePosition pos) {
    Calendar cal = Calendar.getInstance(this.timeZone, this.locale);
    cal.clear();
    return parse(source, pos, cal) ? cal.getTime() : null;
  }
  
  public boolean parse(String source, ParsePosition pos, Calendar calendar) {
    ListIterator<StrategyAndWidth> lt = this.patterns.listIterator();
    while (lt.hasNext()) {
      StrategyAndWidth strategyAndWidth = lt.next();
      int maxWidth = strategyAndWidth.getMaxWidth(lt);
      if (!strategyAndWidth.strategy.parse(this, calendar, source, pos, maxWidth))
        return false; 
    } 
    return true;
  }
  
  private static StringBuilder simpleQuote(StringBuilder sb, String value) {
    for (int i = 0; i < value.length(); i++) {
      char c = value.charAt(i);
      switch (c) {
        case '$':
        case '(':
        case ')':
        case '*':
        case '+':
        case '.':
        case '?':
        case '[':
        case '\\':
        case '^':
        case '{':
        case '|':
          sb.append('\\');
          break;
      } 
      sb.append(c);
    } 
    if (sb.charAt(sb.length() - 1) == '.')
      sb.append('?'); 
    return sb;
  }
  
  private static Map<String, Integer> appendDisplayNames(Calendar calendar, Locale locale, int field, StringBuilder regex) {
    Map<String, Integer> values = new HashMap<>();
    locale = LocaleUtils.toLocale(locale);
    Map<String, Integer> displayNames = calendar.getDisplayNames(field, 0, locale);
    TreeSet<String> sorted = new TreeSet<>(LONGER_FIRST_LOWERCASE);
    for (Map.Entry<String, Integer> displayName : displayNames.entrySet()) {
      String key = ((String)displayName.getKey()).toLowerCase(locale);
      if (sorted.add(key))
        values.put(key, displayName.getValue()); 
    } 
    for (String symbol : sorted)
      simpleQuote(regex, symbol).append('|'); 
    return values;
  }
  
  private int adjustYear(int twoDigitYear) {
    int trial = this.century + twoDigitYear;
    return (twoDigitYear >= this.startYear) ? trial : (trial + 100);
  }
  
  private Strategy getStrategy(char f, int width, Calendar definingCalendar) {
    switch (f) {
      default:
        throw new IllegalArgumentException("Format '" + f + "' not supported");
      case 'D':
        return DAY_OF_YEAR_STRATEGY;
      case 'E':
        return getLocaleSpecificStrategy(7, definingCalendar);
      case 'F':
        return DAY_OF_WEEK_IN_MONTH_STRATEGY;
      case 'G':
        return getLocaleSpecificStrategy(0, definingCalendar);
      case 'H':
        return HOUR_OF_DAY_STRATEGY;
      case 'K':
        return HOUR_STRATEGY;
      case 'M':
        return (width >= 3) ? getLocaleSpecificStrategy(2, definingCalendar) : NUMBER_MONTH_STRATEGY;
      case 'S':
        return MILLISECOND_STRATEGY;
      case 'W':
        return WEEK_OF_MONTH_STRATEGY;
      case 'a':
        return getLocaleSpecificStrategy(9, definingCalendar);
      case 'd':
        return DAY_OF_MONTH_STRATEGY;
      case 'h':
        return HOUR12_STRATEGY;
      case 'k':
        return HOUR24_OF_DAY_STRATEGY;
      case 'm':
        return MINUTE_STRATEGY;
      case 's':
        return SECOND_STRATEGY;
      case 'u':
        return DAY_OF_WEEK_STRATEGY;
      case 'w':
        return WEEK_OF_YEAR_STRATEGY;
      case 'Y':
      case 'y':
        return (width > 2) ? LITERAL_YEAR_STRATEGY : ABBREVIATED_YEAR_STRATEGY;
      case 'X':
        return ISO8601TimeZoneStrategy.getStrategy(width);
      case 'Z':
        if (width == 2)
          return ISO8601TimeZoneStrategy.access$400(); 
        break;
      case 'z':
        break;
    } 
    return getLocaleSpecificStrategy(15, definingCalendar);
  }
  
  private static final ConcurrentMap<Locale, Strategy>[] caches = (ConcurrentMap<Locale, Strategy>[])new ConcurrentMap[17];
  
  private static ConcurrentMap<Locale, Strategy> getCache(int field) {
    synchronized (caches) {
      if (caches[field] == null)
        caches[field] = new ConcurrentHashMap<>(3); 
      return caches[field];
    } 
  }
  
  private Strategy getLocaleSpecificStrategy(int field, Calendar definingCalendar) {
    ConcurrentMap<Locale, Strategy> cache = getCache(field);
    Strategy strategy = cache.get(this.locale);
    if (strategy == null) {
      strategy = (field == 15) ? new TimeZoneStrategy(this.locale) : new CaseInsensitiveTextStrategy(field, definingCalendar, this.locale);
      Strategy inCache = cache.putIfAbsent(this.locale, strategy);
      if (inCache != null)
        return inCache; 
    } 
    return strategy;
  }
  
  private static final Strategy ABBREVIATED_YEAR_STRATEGY = (Strategy)new Object(1);
  
  private static final Strategy NUMBER_MONTH_STRATEGY = (Strategy)new Object(2);
  
  private static final Strategy LITERAL_YEAR_STRATEGY = new NumberStrategy(1);
  
  private static final Strategy WEEK_OF_YEAR_STRATEGY = new NumberStrategy(3);
  
  private static final Strategy WEEK_OF_MONTH_STRATEGY = new NumberStrategy(4);
  
  private static final Strategy DAY_OF_YEAR_STRATEGY = new NumberStrategy(6);
  
  private static final Strategy DAY_OF_MONTH_STRATEGY = new NumberStrategy(5);
  
  private static final Strategy DAY_OF_WEEK_STRATEGY = (Strategy)new Object(7);
  
  private static final Strategy DAY_OF_WEEK_IN_MONTH_STRATEGY = new NumberStrategy(8);
  
  private static final Strategy HOUR_OF_DAY_STRATEGY = new NumberStrategy(11);
  
  private static final Strategy HOUR24_OF_DAY_STRATEGY = (Strategy)new Object(11);
  
  private static final Strategy HOUR12_STRATEGY = (Strategy)new Object(10);
  
  private static final Strategy HOUR_STRATEGY = new NumberStrategy(10);
  
  private static final Strategy MINUTE_STRATEGY = new NumberStrategy(12);
  
  private static final Strategy SECOND_STRATEGY = new NumberStrategy(13);
  
  private static final Strategy MILLISECOND_STRATEGY = new NumberStrategy(14);
}
